"""Views for the ``media_library`` app."""
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import Http404
from django.utils.decorators import method_decorator
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    TemplateView,
    UpdateView,
)
from django.shortcuts import get_object_or_404

from django_libs.views_mixins import AjaxResponseMixin
from multilingual_tags.models import Tag
from user_media.models import UserMediaImage

from . import models
from . import forms


# Mixins ======================================================================

class MediaItemImageCRUDViewMixin(object):
    form_class = forms.MediaItemImageForm
    model = models.MediaItem
    template_name = 'media_library/mediaitem_image_form.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.image = get_object_or_404(
            UserMediaImage, pk=kwargs.get('image_pk'),
        )
        if self.image.user != request.user:
            raise Http404
        return super(MediaItemImageCRUDViewMixin, self).dispatch(
            request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super(MediaItemImageCRUDViewMixin, self).get_form_kwargs()
        kwargs.update({'user': self.request.user, 'image': self.image})
        return kwargs

    def get_context_data(self, **kwargs):
        ctx = super(MediaItemImageCRUDViewMixin, self).get_context_data(
            **kwargs)
        tags = [str(t.name) for t in Tag.objects.get_for_model(self.model)]
        ctx.update({'image': self.image, 'tags': tags})
        return ctx

    def get_success_url(self):
        if 'success_url' in self.request.REQUEST:
            return self.request.REQUEST.get('success_url')
        return reverse('medialibrary_edit')


class MediaItemVideoCRUDViewMixin(CreateView):
    form_class = forms.MediaItemVideoForm
    model = models.MediaItem
    template_name = 'media_library/mediaitem_video_form.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if 'pk' in kwargs:
            self.object = self.get_object(self.queryset)
            if self.object.library.user != request.user:
                raise Http404
        return super(MediaItemVideoCRUDViewMixin, self).dispatch(
            request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(MediaItemVideoCRUDViewMixin, self).get_context_data(
            **kwargs)
        tags = [str(t.name) for t in Tag.objects.get_for_model(self.model)]
        ctx.update({'tags': tags})
        return ctx

    def get_form_kwargs(self):
        kwargs = super(MediaItemVideoCRUDViewMixin, self).get_form_kwargs()
        kwargs.update({'user': self.request.user})
        return kwargs

    def get_success_url(self):
        if 'success_url' in self.request.REQUEST:
            return self.request.REQUEST.get('success_url')
        return reverse('medialibrary_edit')


# Views =======================================================================

class MediaLibraryEditView(AjaxResponseMixin, DetailView):
    """A view to edit media details of a user on its profile."""
    template_name = 'media_library/medialibrary_edit.html'
    model = models.MediaLibrary

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(MediaLibraryEditView, self).dispatch(
            request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(MediaLibraryEditView, self).get_context_data(**kwargs)
        new_images = self.object.user_media_images.filter(
            media_items__isnull=True,
        )
        ctx.update({'new_images': new_images})
        return ctx

    def get_object(self, queryset=None):
        library, created = models.MediaLibrary.objects.get_or_create(
            user=self.request.user
        )
        return library


class MediaItemDeleteView(DeleteView):
    model = models.MediaItem
    template_name = 'media_library/mediaitem_delete.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object(self.queryset)
        if self.object.library.user != request.user:
            raise Http404
        return super(MediaItemDeleteView, self).dispatch(
            request, *args, **kwargs)

    def get_success_url(self):
        if 'success_url' in self.request.REQUEST:
            return self.request.REQUEST.get('success_url')
        return reverse('medialibrary_edit')


class MediaItemImageCreateView(MediaItemImageCRUDViewMixin, CreateView):
    """View to create meta data for a media item."""


class MediaItemImageUpdateView(MediaItemImageCRUDViewMixin, UpdateView):
    """View to update meta data for a media item."""

    def get_object(self, queryset=None):
        return self.image.media_items.all()[0]


class MediaItemVideoCreateView(MediaItemVideoCRUDViewMixin, CreateView):
    """View to create meta data for a media item."""


class MediaItemVideoUpdateView(MediaItemVideoCRUDViewMixin, UpdateView):
    """View to update meta data for a media item."""


class MediaLibraryUploadView(TemplateView):
    """View where users upload their pictures."""
    template_name = "media_library/medialibrary_upload.html"

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(MediaLibraryUploadView, self).dispatch(request, *args,
                                                            **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(MediaLibraryUploadView, self).get_context_data(**kwargs)
        ctx.update({'library': self.request.user.medialibraries.all()[0]})
        return ctx



"""Factories for the models of the ``media_library`` app."""
import factory
from django_libs.tests.factories import HvadFactoryMixin, UserFactory

from media_library import models


class MediaLibraryFactory(factory.DjangoModelFactory):
    """Factory for the ``MediaLibrary`` model class."""
    FACTORY_FOR = models.MediaLibrary

    user = factory.SubFactory(UserFactory)


class MediaItemFactory(HvadFactoryMixin, factory.DjangoModelFactory):
    """Factory for the ``MediaItem`` model class."""
    FACTORY_FOR = models.MediaItem

    library = factory.SubFactory(MediaLibraryFactory)
    video = 'https://youtube.com/watch?v=123456'

    language_code = 'en'



"""Tests for the forms of the ``media_library`` app."""
from django.test import TestCase

from django_libs.tests.factories import UserFactory
from user_media.tests.factories import UserMediaImageFactory
from multilingual_tags.models import Tag, TaggedItem

from . import factories
from .. import forms, models


class MediaItemImageFormTestCase(TestCase):
    """Tests for the ``MediaItemImageForm`` form class."""
    longMessage = True

    def setUp(self):
        self.user = UserFactory()
        self.library = factories.MediaLibraryFactory(user=self.user)
        self.image = UserMediaImageFactory(
            user=self.user, content_object=self.library)
        self.data = {
            'location': u'Berlin',
            'date': u'12/12/2012',
            'tags': u'foo, bar, foobar'
        }
        self.data_update = {
            'location': u'Hamburg',
        }

    def test_form(self):
        form = forms.MediaItemImageForm(
            data=self.data, user=self.user, image=self.image,
        )
        self.assertTrue(form.is_valid(), msg=(
            'The form should be valid. Errors: {0}'.format(form.errors)))

        instance = form.save()
        self.assertEqual(models.MediaItem.objects.count(), 1, msg=(
            'The form should have created one MediaItem.'
        ))
        self.assertEqual(Tag.objects.count(), 3, msg=(
            'The form should have created 3 tags.'
        ))
        self.assertEqual(TaggedItem.objects.count(), 3, msg=(
            'The form should have created 3 tagged items.'
        ))

        form = forms.MediaItemImageForm(
            data=self.data_update, user=self.user, image=self.image,
            instance=instance,
        )
        self.assertTrue(form.is_valid(), msg=(
            'The form should be valid. Errors: {0}'.format(form.errors)))

        instance = form.save()
        self.assertEqual(models.MediaItem.objects.count(), 1, msg=(
            'There should still be only one MediaItem.'
        ))
        self.assertEqual(Tag.objects.count(), 3, msg=(
            'There should still be only 3 tags'
        ))
        self.assertEqual(TaggedItem.objects.count(), 3, msg=(
            'There should still be only 3 tagged items.'
        ))
        self.assertEqual(instance.location, 'Hamburg', msg=(
            'The form should have updated the location field.'
        ))


class MediaItemVideoFormTestCase(TestCase):
    """Tests for the ``MediaItemVideoForm`` form class."""
    longMessage = True

    def setUp(self):
        self.user = UserFactory()
        self.library = factories.MediaLibraryFactory(user=self.user)
        self.data = {
            'video': u'https://youtube.com/watch?v=123456',
            'location': u'Berlin',
            'date': u'12/12/2012',
            'tags': u'foo, bar, foobar'
        }
        self.data_update = {
            'location': u'Hamburg',
        }

    def test_form(self):
        form = forms.MediaItemVideoForm(data=self.data, user=self.user)
        self.assertTrue(form.is_valid(), msg=(
            'The form should be valid. Errors: {0}'.format(form.errors)))

        instance = form.save()
        self.assertEqual(models.MediaItem.objects.count(), 1, msg=(
            'The form should have created one MediaItem.'
        ))

        form = forms.MediaItemVideoForm(
            data=self.data_update, user=self.user, instance=instance,
        )
        self.assertTrue(form.is_valid(), msg=(
            'The form should be valid. Errors: {0}'.format(form.errors)))

        form.save()
        self.assertEqual(models.MediaItem.objects.count(), 1, msg=(
            'There should still be only one MediaItem.'
        ))



"""Tests for the utils of the ``media_library`` app."""
from django.test import TestCase

from .. import utils


class GetVideoIDTestCase(TestCase):
    """Tests for the ``get_video_id`` function."""
    longMessage = True

    def test_function(self):
        self.assertEqual(utils.get_video_id('vimeo.com/12345'), '12345')
        self.assertEqual(utils.get_video_id('vimeo.com/12345/'), '12345')
        self.assertEqual(
            utils.get_video_id('vimeo.com/username/12345'), '12345')
        self.assertEqual(
            utils.get_video_id('vimeo.com/groupname/username/12345'), '12345')
        self.assertEqual(
            utils.get_video_id('youtube.com/watch?v=12345'), '12345')
        self.assertEqual(
            utils.get_video_id('youtube.com/watch?v=12345&list=abcd#t=21'),
            '12345')


class ValidateVideoURLTestCase(TestCase):
    """Tests for the ``validate_video_url`` function."""

    def test_function(self):
        self.yt_links = [
            'http://www.youtube.com/watch?v=-JyZLS2IhkQ',
            'https://www.youtube.com/watch?v=-JyZLS2IhkQ',
            'http://www.youtube.de/watch?v=-JyZLS2IhkQ&list=foo',
            'http://www.youtube.de/watch?v=-JyZLS2IhkQ',
            'https://youtube.com/watch?v=-JyZLS2IhkQ',
            ('https://www.youtube.com/watch?v=PguLNvCcOHQ'
             '&list=RDPguLNvCcOHQ#t=0'),
            'http://youtu.be/PguLNvCcOHQ',
            'http://youtu.be/PguLNvCcOHQ?list=RDPguLNvCcOHQ',
        ]
        self.vm_links = [
            'http://vimeo.com/channels/staffpicks/110140870',
            'http://vimeo.com/59777392',
            'http://vimeopro.com/foo/bar/110140870',
            'http://vimeo.com/video/59777392',
            'http://www.vimeo.com/video/59777392',
            ('http://vimeo.com/groups/thedirectorofphotography/'
             'videos/110016243'),
        ]
        self.invalid_links = [
            'http://vimeo.com/',
            'http://vimeo.com/groups/thedirectorofphotography/',
            'http://vimeo.com/channels/staffpicks/',
            'http://www.youtube.de/watch?v=',
            'https://www.youtube.com/playlist?list=PL65E33789AA7052BC',
            'http://example.com/',
        ]
        for url in self.yt_links:
            result = utils.validate_video_url(url)
            self.assertEqual(
                result,
                'youtube',
                msg=('Invalid result "{0}" for url "{1}".'.format(
                    result, url)),
            )
        for url in self.vm_links:
            result = utils.validate_video_url(url)
            self.assertEqual(
                result,
                'vimeo',
                msg=('Invalid result "{0}" for url "{1}".'.format(
                    result, url)),
            )
        for url in self.invalid_links:
            result = utils.validate_video_url(url)
            self.assertIsNone(
                result,
                'Result should be "None" and not "{0}" for "{1}"'.format(
                    result, url
                )
            )



"""Utilities for the ``media_library`` app."""
import re
VIMEO_PATTERN = re.compile('(?:https?://)?(?:www.)?(vimeo)(?:pro)?.com/(?:.*/)?(\d+)')
YOUTUBE_PATTERN = re.compile(
    '(?:https?://)?(?:(?:www\.)?(youtube)(?:\.\w{2,3})+/watch\?.*v=([a-zA-Z0-9_-]+)&?.*|(youtu\.be)/([a-zA-Z0-9_-]+))'  # NOQA
)
VIDEO_PATTERNS = [VIMEO_PATTERN, YOUTUBE_PATTERN]


def get_video_id(url):
    """Returns the video ID from a vimeo or YouTube URL."""
    video_id = ''
    if not url:
        return ''
    if validate_video_url(url) == 'vimeo':
        video_id = VIMEO_PATTERN.findall(url)[0][1]
    elif validate_video_url(url) == 'youtube':
        video_id = YOUTUBE_PATTERN.findall(url)[0][1]
    return video_id


def validate_video_url(url):
    """Validates, where the video was from."""
    for pattern in VIDEO_PATTERNS:
        result = pattern.search(url)
        if result is not None:
            # cast to list and remove NoneType values
            result = [res for res in result.groups() if res is not None]
            if len(result) == 2 and result[1] is not None:
                origin = result[0]
                if origin.startswith('youtu'):
                    return 'youtube'
                else:
                    return origin
    return



"""Models for the ``media_library`` app."""
from django.contrib.contenttypes import generic
from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from hvad.models import TranslatableModel, TranslatedFields

from .utils import get_video_id, validate_video_url


class MediaLibrary(models.Model):
    """
    General collection of ``MediaItem`` objects for a user.

    :user: FK to the ``User`` object this collection belongs to.
    :user_media_images: a generic relation to user media images of this user

    """

    user = models.ForeignKey(
        'auth.User',
        verbose_name=_('user'),
        related_name='medialibraries',
    )

    user_media_images = generic.GenericRelation(
        'user_media.UserMediaImage',
    )

    def __unicode__(self):
        return 'Library of "{0}"'.format(self.user.email)

    def get_absolute_url(self):
        return reverse('showreel_edit')


class MediaItem(TranslatableModel):
    """
    Stores video URLs and references ``UserMediaImage`` and extends them with
    some additional meta data.

    :library: the library this item belongs to.
    :image: FK to a ``UserMediaImage``. Optional.
    :video: URL of a youtube video. Optional
    :date: The date this video or image was taken.
    :tags: A generic relation to the ``TaggedItem`` from ``multilingual_tags``.
    :generic_position: A generic relation to ``ObjectPosition``.

    translatable:
    :location: The place where this image or video was taken

    """

    library = models.ForeignKey(
        MediaLibrary,
        verbose_name=_('library'),
        related_name='media_items',
    )

    image = models.ForeignKey(
        'user_media.UserMediaImage',
        verbose_name=_('image'),
        related_name='media_items',
        blank=True, null=True,
    )

    video = models.URLField(
        verbose_name=_('video'),
        help_text=_('Enter a YouTube or Vimeo video URL here.'),
        max_length=512,
        blank=True, null=True,
    )

    date = models.DateField(
        verbose_name=_('date'),
        blank=True, null=True,
    )

    translations = TranslatedFields(
        location=models.CharField(
            verbose_name=_('location'),
            max_length=256,
            blank=True,
        )
    )

    tags = generic.GenericRelation(
        'multilingual_tags.TaggedItem',
    )

    generic_position = generic.GenericRelation(
        'generic_positions.ObjectPosition',
    )

    def __unicode__(self):
        if self.image is not None:
            return 'Image of "{0}"'.format(self.library.user)
        return 'Video of "{0}"'.format(self.library.user)

    def clean(self):
        """
        Checks if the video has the correct format.

        Currently supported are vimeo and youtube.

        """
        if self.video and self.get_video_origin() is None:
            raise ValidationError(_(
                'The video URL was not in a valid Vimeo or YouTube format.'
            ))

    def get_video_origin(self):
        return validate_video_url(self.video)

    def get_user(self):
        """Returns a user for the ``django-multilingual-tags`` form API."""
        return self.library.user

    def delete(self, using=None):
        # making sure, the UserMediaImage instance is removed as well.
        if self.image is not None:
            self.image.delete()
        super(MediaItem, self).delete(using=using)

    @property
    def video_id(self):
        """Returns the video ID from the URL."""
        return get_video_id(self.video)
